@@ -9,18 +9,27 @@ import re
 from datetime import datetime
 from pathlib import Path
 
-import cchardet
 import click
+import cjkwrap
 from bs4 import UnicodeDammit
 from ebooklib import epub
+from langdetect import detect
+
+from txt2ebook import (
+    __version__,
+    IDEOGRAPHIC_SPACE,
+    SPACE,
+    ZH_NUMS_WORDS,
+    ZH_FULLWIDTH_NUMS
+)
 
-from txt2ebook import __version__
 
 logger = logging.getLogger(__name__)
 
 
 @click.command(no_args_is_help=True)
-@click.argument("filename", type=click.Path(exists=True))
+@click.argument("input_file", type=click.Path(exists=True))
+@click.argument("output_file", type=click.Path(), required=False)
 @click.option(
     "--title",
     "-t",
@@ -31,15 +40,13 @@ logger = logging.getLogger(__name__)
 @click.option(
     "--language",
     "-l",
-    default="en",
-    show_default=True,
+    default=None,
     help="Set the language of the ebook.",
 )
 @click.option(
     "--author",
     "-a",
     default=None,
-    show_default=True,
     help="Set the author of the ebook.",
 )
 @click.option(
@@ -47,7 +54,6 @@ logger = logging.getLogger(__name__)
     "-c",
     type=click.Path(exists=True),
     default=None,
-    show_default=True,
     help="Set the cover of the ebook.",
 )
 @click.option(
@@ -67,12 +73,19 @@ logger = logging.getLogger(__name__)
     help="Do not backup source txt file.",
 )
 @click.option(
-    "--remove_wrapping",
-    "-rw",
+    "--no-wrapping",
+    "-nw",
     is_flag=True,
     show_default=True,
     help="Remove word wrapping.",
 )
+@click.option(
+    "--width",
+    "-w",
+    type=click.INT,
+    show_default=True,
+    help="Set the width for line wrapping.",
+)
 @click.option(
     "--delete-regex",
     "-dr",
@@ -93,8 +106,7 @@ logger = logging.getLogger(__name__)
     help="Regex to delete whole line.",
 )
 @click.version_option(prog_name="txt2ebook", version=__version__)
-@click.pass_context
-def main(ctx, **kwargs):
+def main(**kwargs):
     """
     Console tool to convert txt file to different ebook format.
     """
@@ -104,12 +116,8 @@ def main(ctx, **kwargs):
         datefmt="%Y-%m-%d %H:%M:%S",
     )
 
-    ctx.obj = kwargs
-
     try:
-        filename = Path(kwargs["filename"])
-        title = kwargs["title"]
-
+        filename = Path(kwargs["input_file"])
         logger.info("Processing txt file: '%s'.", filename.resolve())
 
         with open(filename, "rb") as file:
@@ -120,27 +128,49 @@ def main(ctx, **kwargs):
             if not content:
                 raise RuntimeError(f"Empty file content in '{filename}'!")
 
-            content = tidy_content(content)
+            kwargs["language"] = kwargs["language"] or detect_language(content)
+            kwargs["author"] = kwargs["author"] or detect_author(content)
+            kwargs["title"] = kwargs["title"] or detect_book_title(content)
+            output_filename = Path(
+                kwargs["output_file"]
+                or Path(kwargs["title"] or filename).stem + ".epub"
+            )
+
+            if kwargs["delete_regex"]:
+                content = do_delete_regex(content, kwargs)
+
+            if kwargs["replace_regex"]:
+                content = do_replace_regex(content, kwargs)
+
+            if kwargs["delete_line_regex"]:
+                content = do_delete_regex(content, kwargs)
+
+            if kwargs["no_wrapping"]:
+                content = do_no_wrapping(content)
+
+            if kwargs["width"]:
+                content = do_wrapping(content, kwargs)
+
             chapters = parse_content(content)
 
-        ctx.obj["title"] = ctx.obj["title"] or find_book_title(content)
-        output_filename = ctx.obj["title"] or Path(filename).stem
-
-        if not ctx.obj["author"]:
-            match = re.search(r"作者：(.*)", content)
-            if match:
-                author = match.group(1)
-                ctx.obj["author"] = author
-
         build_txt(filename, content, kwargs["no_backup"])
         if chapters:
-            build_epub(output_filename, chapters)
+            build_epub(output_filename, chapters, kwargs)
 
     except RuntimeError as error:
         click.echo(f"[E] {str(error)}!", err=True)
 
 
-def find_book_title(content):
+def detect_language(content):
+    """
+    Detect the language (ISO 639-1) of the content of the txt file.
+    """
+    language = detect(content)
+    logger.info("Language detected: '%s'.", language)
+    return language
+
+
+def detect_book_title(content):
     """
     Extract book title from the content of the txt file.
     """
@@ -151,117 +181,134 @@ def find_book_title(content):
             (title for title in match.groups() if title is not None)
         )
         logger.info("Found book title: '%s'.", book_title)
-    else:
-        book_title = False
-        logger.info("No book title found from file!")
-    return book_title
+        return book_title
+
+    logger.info("No book title found from file!")
+    return False
 
 
-@click.pass_obj
-def tidy_content(ctx, content):
+def detect_author(content):
     """
-    Tidy content with incorrect wrapping which can affect chapters splitting
-    and other opinionated changes.
+    Extract author from the content of the txt file.
     """
-    # Remove words/phrases based on regex.
-    for delete_regex in ctx["delete_regex"]:
+    match = re.search(r"作者：(.*)", content)
+    if match:
+        author = match.group(1)
+        logger.info("Found author: '%s'.", author)
+        return author
+
+    logger.info("No author found from file!")
+    return False
+
+
+def do_delete_regex(content, kwargs):
+    """
+    Remove words/phrases based on regex.
+    """
+    for delete_regex in kwargs["delete_regex"]:
         content = re.sub(
             re.compile(rf"{delete_regex}", re.MULTILINE), "", content
         )
+    return content
 
-    # Replace words/phrases based on regex.
-    for search, replace in ctx["replace_regex"]:
+
+def do_replace_regex(content, kwargs):
+    """
+    Replace words/phrases based on regex.
+    """
+    for search, replace in kwargs["replace_regex"]:
         content = re.sub(
             re.compile(rf"{search}", re.MULTILINE), rf"{replace}", content
         )
+    return content
 
-    # Delete whole line based on regex.
-    for delete_line_regex in ctx["delete_line_regex"]:
+
+def do_delete_line_regex(content, kwargs):
+    """
+    Delete whole line based on regex.
+    """
+    for delete_line_regex in kwargs["delete_line_regex"]:
         content = re.sub(
             re.compile(rf"^.*{delete_line_regex}.*$", re.MULTILINE),
             "",
             content,
         )
+    return content
 
-    # Stick to one form of quotation punctuation.
-    content = content.replace("“", "「").replace("”", "」")
 
-    # Remove incorrect closing quotation line wrapping.
-    content = re.sub(r"\s+」", "」\n\n", content)
+def do_no_wrapping(content):
+    """
+    Remove wrapping. Paragraph should be in one line.
+    """
+    # Convert to single spacing before we removed wrapping.
+    lines = content.split("\n")
+    content = "\n\n".join([line.strip() for line in lines if line])
 
-    # # Remove spaces or tabs at beginning of line.
-    # content = re.sub(re.compile(r"^[ \t]+", re.MULTILINE), "", content)
-
-    # # Replace double spaces as single space.
-    # content = content.replace("  ", " ")
-
-    # # Replace extra space in paragraph.
-    # content = re.sub(re.compile(r"(?<![章卷])　", re.MULTILINE), "", content)
-
-    # Append space to chapter name and chapter title.
-    pattern = re.compile(
-        r"^((第[0-9零一二三四五六七八九十百千两]*[章卷]|楔子)(?!\s)[：-]?)(.*)$", re.MULTILINE
-    )
-    chapters = re.findall(pattern, content)
-    for chapter in chapters:
-        space = " "
-        content = content.replace(
-            f"{chapter[0]}{chapter[2]}", f"{chapter[1]}{space}{chapter[2]}", 1
-        )
-
-    # Remove duplicate chapter title.
-    pattern = re.compile(rf"({chapter_regexs()})([\r?\n]+\1)$", re.MULTILINE)
-    duplicates = re.findall(pattern, content)
-    for duplicate in duplicates:
-        content = content.replace(duplicate[1], "", 1)
-
-    # # Single spacing.
-    # lines = content.split("\n")
-    # content = "\n\n".join([line.strip() for line in lines if line])
-
-    # Remove wrapping. Paragraph should be in one line.
-    if ctx["remove_wrapping"]:
-        unwrapped_content = ""
-        for line in content.split("\n\n"):
-            # if a line contains more opening quote(「) than closing quote(」),
-            # we're still within the same paragraph.
-            # e.g.:
-            # 「...」「...
-            # 「...
-            if line.count("「") > line.count("」"):
-                unwrapped_content = unwrapped_content + line.strip()
-            elif (
-                re.search(r"[…。？！]{1}」?$", line)
-                or re.search(r"」$", line)
-                or re.match(r"^[ \t]*……[ \t]*$", line)
-                or re.match(r"^「」$", line)
-                or re.match(r".*[》：＊\*]$", line)
-                or re.match(r".*[a-zA-Z0-9]$", line)
-            ):
-                unwrapped_content = unwrapped_content + line.strip() + "\n\n"
-            elif re.match(chapter_regexs(), line):
-                # replace full-width space with half-wdith space.
-                # looks nicer on the output.
-                header = line.replace("\u3000\u3000", " ").replace(
-                    "\u3000", " "
-                )
-                unwrapped_content = (
-                    unwrapped_content + "\n\n" + header.strip() + "\n\n"
-                )
-            else:
-                unwrapped_content = unwrapped_content + line.strip()
-    else:
-        unwrapped_content = content
+    unwrapped_content = ""
+    for line in content.split("\n\n"):
+        # if a line contains more opening quote(「) than closing quote(」),
+        # we're still within the same paragraph.
+        # e.g.:
+        # 「...」「...
+        # 「...
+        if line.count("「") > line.count("」"):
+            unwrapped_content = unwrapped_content + line.strip()
+        elif (
+            re.search(r"[…。？！]{1}」?$", line)
+            or re.search(r"」$", line)
+            or re.match(r"^[ \t]*……[ \t]*$", line)
+            or re.match(r"^「」$", line)
+            or re.match(r".*[》：＊\*]$", line)
+            or re.match(r".*[a-zA-Z0-9]$", line)
+        ):
+            unwrapped_content = unwrapped_content + line.strip() + "\n\n"
+        elif re.match(chapter_regexs(), line):
+            # replace full-width space with half-wdith space.
+            # looks nicer on the output.
+            header = line.replace(IDEOGRAPHIC_SPACE*2, SPACE).replace(
+                IDEOGRAPHIC_SPACE, SPACE
+            )
+            unwrapped_content = (
+                unwrapped_content + "\n\n" + header.strip() + "\n\n"
+            )
+        else:
+            unwrapped_content = unwrapped_content + line.strip()
 
     return unwrapped_content
 
 
+def do_wrapping(content, kwargs):
+    """
+    Wrapping and filling CJK text.
+    """
+    logger.info("Wrapping paragraph to width: %s.", kwargs["width"])
+
+    paragraphs = []
+    # We don't remove empty line and keep all formatting as it.
+    for paragraph in content.split("\n"):
+        paragraph = paragraph.strip()
+
+        lines = cjkwrap.wrap(paragraph, width=kwargs["width"])
+        paragraph = "\n".join(lines)
+        paragraphs.append(paragraph)
+
+    wrapped_content = "\n".join(paragraphs)
+    return wrapped_content
+
+
+# # Single spacing.
+# lines = content.split("\n")
+# content = "\n\n".join([line.strip() for line in lines if line])
+
+
 def parse_content(content):
     """
     Parse the content into volumes (if exists) and chapters.
     """
 
-    volume_regex = "^[ \t]*第[0-9一二三四五六七八九十]*[集卷册][^。~\n]*$"
+    spaces = f"[{SPACE}\t{IDEOGRAPHIC_SPACE}]"
+    volume_seq = f"[0-9{ZH_FULLWIDTH_NUMS}{ZH_NUMS_WORDS}]"
+    volume_regex = f"^{spaces}*第{volume_seq}*[集卷册][^。~\n]*$"
     volume_pattern = re.compile(rf"{volume_regex}", re.MULTILINE)
     volume_headers = re.findall(volume_pattern, content)
 
@@ -300,25 +347,28 @@ def parse_chapters(content):
 
     if not chapter_headers:
         return False
-    else:
-        bodies = re.split(chapter_pattern, content)
-        chapters = list(zip(chapter_headers, bodies[1:]))
 
-        return chapters
+    bodies = re.split(chapter_pattern, content)
+    chapters = list(zip(chapter_headers, bodies[1:]))
+
+    return chapters
 
 
 def chapter_regexs():
     """
     Regex rules for chapter headers.
     """
+    spaces = f"[{SPACE}\t{IDEOGRAPHIC_SPACE}]"
+    chapter_seq = f"[.0-9{ZH_FULLWIDTH_NUMS}{ZH_NUMS_WORDS}]"
+
     regexs = (
-        "^[ \t]*第[.0-9零一二三四五六七八九十百千两]*[章篇折][^。~\n]*$",
-        "^[ \t]*[楔引]子.*$",
-        "^[ \t]*序[章幕曲]?.*$",
-        "^[ \t]*前言.*$",
-        "^[ \t]*[内容]*简介.*$",
-        "^[ \t]*[号番]外篇.*$",
-        "^[ \t]*尾声$",
+        f"^{spaces}*第{chapter_seq}*[章篇回折][^。\n]*$",
+        f"^{spaces}*[楔引]子[^，].*$",
+        f"^{spaces}*序[章幕曲]?.*$",
+        f"^{spaces}*前言.*$",
+        f"^{spaces}*[内容]*简介.*$",
+        f"^{spaces}*[号番]外篇.*$",
+        f"^{spaces}*尾声$",
     )
     return "|".join(regexs)
 
@@ -344,26 +394,26 @@ def build_txt(filename, parsed_content, no_backup):
         logger.info("Overwrite txt file: '%s'.", txt_filename.resolve())
 
 
-@click.pass_obj
-def build_epub(ctx, output_filename, parsed_content):
+def build_epub(output_filename, parsed_content, kwargs):
     """
     Generate epub from the parsed chapters from txt file.
     """
 
     book = epub.EpubBook()
 
-    if ctx["title"]:
-        book.set_title(ctx["title"])
+    if kwargs["title"]:
+        book.set_title(kwargs["title"])
 
-    if ctx["language"]:
-        book.set_language(ctx["language"])
+    if kwargs["language"]:
+        book.set_language(kwargs["language"])
 
-    if ctx["author"]:
-        book.add_author(ctx["author"])
+    if kwargs["author"]:
+        book.add_author(kwargs["author"])
 
-    if ctx["cover"]:
-        book.set_cover("cover.jpg", open(ctx["cover"], "rb").read())
-        book.spine += ["cover"]
+    if kwargs["cover"]:
+        with open(kwargs["cover"], "rb") as image:
+            book.set_cover("cover.jpg", image.read())
+            book.spine += ["cover"]
 
     book.add_item(epub.EpubNcx())
     book.add_item(epub.EpubNav())
@@ -387,8 +437,9 @@ def build_epub(ctx, output_filename, parsed_content):
             book.spine += [html_chapter]
             book.toc += [html_chapter]
 
-    logger.info("Generating epub file: '%s'.", output_filename + ".epub")
-    epub.write_epub(output_filename + ".epub", book, {})
+    output_filename.parent.mkdir(parents=True, exist_ok=True)
+    epub.write_epub(output_filename, book, {})
+    logger.info("Generating epub file: '%s'.", output_filename)
 
 
 def build_html_chapter(title, body, volume=None):
@@ -397,16 +448,16 @@ def build_html_chapter(title, body, volume=None):
     """
     if volume:
         filename = f"{volume}_{title}"
-        logger.debug(f"  {title}")
+        logger.debug("%s%s" % SPACE*2, title)
     else:
         filename = title
         logger.debug(title)
 
-    filename = filename.replace(" ", "_")
+    filename = filename.replace(SPACE, "_")
 
     html = f"<h2>{title}</h2>"
     for paragraph in body.split("\n\n"):
-        paragraph = paragraph.replace(" ", "").replace("\n", "")
+        paragraph = paragraph.replace(SPACE, "").replace("\n", "")
         html = html + f"<p>{paragraph}</p>"
 
     return epub.EpubHtml(
